#include <wdm.h>
#include <stdarg.h>
#include <stdio.h>
#include <usbdi.h>
#include <usbdlib.h>

#include "CanUsbInternal.h"

//returns the port status for our device
//return value: STATUS_SUCCESS if successful, STATUS_UNSUCCESSFUL otherwise
NTSTATUS CanUsbGetPortStatus( IN PDEVICE_OBJECT pDeviceObject, IN PULONG pPortStatus )
{
	NTSTATUS WaitStatus, Status = STATUS_SUCCESS;
	PIRP pIrp;
	KEVENT evEvent;
	IO_STATUS_BLOCK IoStatus;
	PIO_STACK_LOCATION pNextStack;
	PDEVICE_EXTENSION pDeviceExtension;

	CanUsbDebug( DBGLVL_DEFAULT, ( "entering CanUsbGetPortStatus\r\n" ) );

	pDeviceExtension = pDeviceObject->DeviceExtension;

	*pPortStatus = 0;

	//issue a synchronous request
	KeInitializeEvent( &evEvent, NotificationEvent, FALSE );

	//IoBuildDeviceIoControlRequest allocates and sets up an IRP for a device control request
	pIrp = IoBuildDeviceIoControlRequest( IOCTL_INTERNAL_USB_GET_PORT_STATUS,
			pDeviceExtension->pTopOfStackDeviceObject, //next-lower driver's device object, representing the target device.
			NULL, 0, NULL, 0, //no input or output buffers
			TRUE, //internal (use IRP_MJ_INTERNAL_DEVICE_CONTROL)
			&evEvent, //event to be signalled on completion (we wait for it below)
			&IoStatus );

	//Call the class driver to perform the operation.
	//If the returned status is PENDING, wait for the request to complete.
	//IoGetNextIrpStackLocation gives a higher level driver access to the next-lower 
	//driver's I/O stack location in an IRP so the caller can set it up for the lower driver.
	pNextStack = IoGetNextIrpStackLocation( pIrp );
	CanUsbAssert( pNextStack != NULL );
	pNextStack->Parameters.Others.Argument1 = pPortStatus;

	CanUsbDebug( DBGLVL_DEFAULT, ( "CanUsbGetPortStatus() calling USBD port status api\r\n" ) );

	Status = IoCallDriver( pDeviceExtension->pTopOfStackDeviceObject, pIrp );

	CanUsbDebug( DBGLVL_DEFAULT, ( "CanUsbGetPortStatus() return from IoCallDriver USBD 0x%x\r\n", Status ) );

	if( Status == STATUS_PENDING )
	{
		CanUsbDebug( DBGLVL_DEFAULT, ( "Wait for single object\r\n" ) );
		WaitStatus = KeWaitForSingleObject( &evEvent, Suspended, KernelMode, FALSE, NULL );
		CanUsbDebug( DBGLVL_DEFAULT, ( "CanUsbGetPortStatus() Wait for single object, returned 0x%x\r\n", WaitStatus ) );
	}
	else
		IoStatus.Status = Status;

	CanUsbDebug( DBGLVL_DEFAULT, ( "CanUsbGetPortStatus() Port status = 0x%x\r\n", *pPortStatus ) );

	//USBD maps the error code for us
	Status = IoStatus.Status;

	CanUsbDebug( DBGLVL_DEFAULT, ( "exiting CanUsbGetPortStatus with status 0x%x\r\n", Status ) );
	return Status;
}

//reset the our parent port
//return value: STATUS_SUCCESS if successful, STATUS_UNSUCCESSFUL otherwise
NTSTATUS CanUsbResetParentPort( IN IN PDEVICE_OBJECT pDeviceObject )
{
	NTSTATUS WaitStatus, Status = STATUS_SUCCESS;
	PIRP pIrp;
	KEVENT evEvent;
	IO_STATUS_BLOCK IoStatus;
	PIO_STACK_LOCATION pNextStack;
	PDEVICE_EXTENSION pDeviceExtension;

	CanUsbDebug( DBGLVL_DEFAULT, ( "entering CanUsbResetParentPort\r\n" ) );

	pDeviceExtension = pDeviceObject->DeviceExtension;

	//issue a synchronous request
	KeInitializeEvent( &evEvent, NotificationEvent, FALSE );
	pIrp = IoBuildDeviceIoControlRequest( IOCTL_INTERNAL_USB_RESET_PORT,
				pDeviceExtension->pTopOfStackDeviceObject,
				NULL, 0, NULL, 0,
				TRUE, //internal (use IRP_MJ_INTERNAL_DEVICE_CONTROL)
				&evEvent, &IoStatus);

	//Call the class driver to perform the operation.
	//If the returned status is PENDING, wait for the request to complete.
	pNextStack = IoGetNextIrpStackLocation( pIrp );
	CanUsbAssert( pNextStack != NULL );

	CanUsbDebug( DBGLVL_DEFAULT, ( "CanUsbResetParentPort() calling USBD enable port api\r\n" ) );

	Status = IoCallDriver( pDeviceExtension->pTopOfStackDeviceObject, pIrp );

	CanUsbDebug( DBGLVL_DEFAULT, ( "CanUsbResetParentPort() return from IoCallDriver USBD 0x%x\r\n", Status ) );

	if( Status == STATUS_PENDING )
	{
		CanUsbDebug( DBGLVL_DEFAULT, ( "CanUsbResetParentPort() Wait for single object\r\n" ) );
		WaitStatus = KeWaitForSingleObject( &evEvent, Suspended, KernelMode, FALSE, NULL );
		CanUsbDebug( DBGLVL_DEFAULT, ( "CanUsbResetParentPort() Wait for single object, returned 0x%x\r\n", WaitStatus ) );
	}
	else
		IoStatus.Status = Status;

	//USBD maps the error code for us
	Status = IoStatus.Status;

	CanUsbDebug( DBGLVL_DEFAULT, ( "exiting BulkUsb_ResetPort with status 0x%x\r\n", Status ) );
	return Status;
}

//Checks port status; if OK, return success and  do no more; If bad, attempt reset
//pDeviceObject: pointer to the device object for this instance of the device
//return value: NT status code
NTSTATUS CanUsbResetDevice( IN PDEVICE_OBJECT pDeviceObject )
{
	NTSTATUS Status;
	ULONG PortStatus;

	CanUsbDebug( DBGLVL_MAXIMUM, ( "entering CanUsbResetDevice()\r\n" ) );
	
	//check the port state, if it is disabled we will need to re-enable it
	Status = CanUsbGetPortStatus( pDeviceObject, &PortStatus );

	if( NT_SUCCESS( Status ) && !(PortStatus & USBD_PORT_ENABLED) && PortStatus & USBD_PORT_CONNECTED)
	{
		//port is disabled, attempt reset
		CanUsbDebug( DBGLVL_DEFAULT, ( "CanUsbResetDevice() will reset\r\n" ) );
		Status = CanUsbResetParentPort( pDeviceObject );
	}
	return Status;
}

//Dispatch table handler for IRP_MJ_DEVICE_CONTROL; 
//Handle DeviceIoControl() calls  from User mode
//pDeviceObject: pointer to the FDO for this instance of the device
//return value: NT status code
NTSTATUS CanUsbIoctl( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp )
{
	PIO_STACK_LOCATION pIrpStack;
	PVOID pIoBuffer;
	int InputBufferLength, OutputBufferLength, OutputLength, ReturnValue;
	PDEVICE_EXTENSION pDeviceExtension;
	ULONG IoControlCode;
	PFILE_OBJECT pFileObject;
	NTSTATUS Status;

	CanUsbDebug( DBGLVL_DEFAULT, ( "IRP_MJ_DEVICE_CONTROL\r\n" ) );

	CanUsbIncIoCount( pDeviceObject );

	//Get a pointer to the current location in the Irp. This is where
	//the function codes and parameters are located.

	pDeviceExtension = pDeviceObject->DeviceExtension;

	//Can't accept a new io request if:
	// 1) device is removed, 
	// 2) has never been started, 
	// 3) is stopped,
	// 4) has a remove request pending,
	// 5) has a stop device pending
	if( !CanUsbCanAcceptIoRequests( pDeviceObject ) )
	{
		Status = STATUS_DELETE_PENDING;
		pIrp->IoStatus.Status = Status;
		pIrp->IoStatus.Information = 0;

		IoCompleteRequest( pIrp, IO_NO_INCREMENT );

		CanUsbDecIoCount( pDeviceObject );
		return Status;
	}

	pIrpStack = IoGetCurrentIrpStackLocation( pIrp );

	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;

	//get pointers and lengths of the caller's (user's) IO buffer
	pIoBuffer = pIrp->AssociatedIrp.SystemBuffer;
	InputBufferLength = pIrpStack->Parameters.DeviceIoControl.InputBufferLength;
	OutputBufferLength = pIrpStack->Parameters.DeviceIoControl.OutputBufferLength;
	IoControlCode = pIrpStack->Parameters.DeviceIoControl.IoControlCode;
	pFileObject = pIrpStack->FileObject;

	//handle IOCTLs from user mode
	if( CanUsbFileIoctl( pFileObject, IoControlCode, pIoBuffer, InputBufferLength, OutputBufferLength, &OutputLength, &ReturnValue ) )
	{
		//success
		Status = STATUS_SUCCESS;
		pIrp->IoStatus.Status = STATUS_SUCCESS;
		//IOCTL succeeded
		if( ReturnValue == 0 )
		{
			//set length of returned data
			pIrp->IoStatus.Information = OutputLength;
		}
		//IOCTL did not succeed
		else
		{
			//put return value in output buffer if it is large enough
			//(we can do this, because all IOCTLs have got an output buffer larger than an int)
			if( OutputBufferLength >= sizeof( int ) )
			{
				*((int *)pIoBuffer) = ReturnValue;
				pIrp->IoStatus.Information = sizeof( int );
			}
			else
				pIrp->IoStatus.Information = 0;
		}
	}
	else
	{
		//error
		Status = STATUS_UNSUCCESSFUL;
		pIrp->IoStatus.Status = STATUS_UNSUCCESSFUL;
		//set length of returned data to zero
		pIrp->IoStatus.Information = 0;
	}

	IoCompleteRequest( pIrp, IO_NO_INCREMENT );

	CanUsbDecIoCount( pDeviceObject );

	return Status;
}
